import os
import subprocess
import openai

class TextToSpeechAction:
    def __init__(self, agent):
        self.agent = agent
        self.name = 'text_to_speech'
        self.description = 'Converts text to speech and returns the path to the generated audio file.'
        self.parameters = [
            {
                'name': 'text',
                'type': 'string',
                'description': 'The text to be converted to speech.',
                'required': True
            },
            {
                'name': 'play',
                'type': 'boolean',
                'description': 'Play the audio file after it is generated.',
                'required': True,
                'default': True
            }
        ]

    async def run(self, args):
        text = args.get('text')
        play= args.get('play', True)
        if os.name == 'posix' and os.uname().sysname == 'Darwin':
            # For macOS, use 'say' command
            subprocess.run(['say', text])
            return f"Text spoken on macOS using Siri: {text}"
        else:
            try:
                
                speech_response = openai.Audio.speech.create(
                    model="tts-1",
                    voice="alloy",
                    input=text
                )
                audio_buffer = speech_response['data']
                audio_file_path = 'speak.mp3'
                with open(audio_file_path, 'wb') as audio_file:
                    audio_file.write(audio_buffer)

                # Play the audio file if required
                if play:
                    if os.name == 'posix':
                        subprocess.run(['play', audio_file_path])
                    elif os.name == 'nt':
                        os.startfile(audio_file_path)

                return f"Text spoken using OpenAI and saved to: {audio_file_path}"
            except Exception as error:
                return f"Error in Text to Speech: {str(error)}"



import asyncio
import json
from aiohttp import web
import socketio

class WebsocketAction:
    def __init__(self, agent):
        self.agent = agent
        self.name = "websocket_server"
        self.description = "Starts a websocket server for real-time communication with the agent."
        self.parameters = [
            {
                "name": "htmlContent",
                "description": "HTML content to be served on the root path",
                "required": True,
                "type": "string",
                "default": "<html><body><h1>Default HTML Content</h1></body></html>"
            }
        ]
        self.app = web.Application()
        self.sio = socketio.AsyncServer(cors_allowed_origins='*')
        self.sio.attach(self.app)

    async def emit_response(self, sid, result):
        await self.sio.emit('agent_response', result, to=sid)

    async def async_agent_interact(self, data):
        data_dict = json.loads(data)
        self.agent.messages = data_dict  # Update the agent's state with the parsed data    
        return await self.agent.interact(True)

    async def index(self, request):
        html_content = self.parameters[0]["default"]  # Default HTML content
        return web.Response(text=html_content, content_type='text/html')

    async def run(self, args):
        # Update HTML content if provided in args
        if "htmlContent" in args:
            self.parameters[0]["default"] = args["htmlContent"]

        self.app.router.add_get('/', self.index)

        @self.sio.event
        async def connect(sid, environ):
            print("A user connected", sid)

        @self.sio.event
        async def disconnect(sid):
            print("A user disconnected", sid)

        @self.sio.event
        async def agent_request(sid, data):
            print("Agent request received", data)
            result = await self.async_agent_interact(data)
            await self.emit_response(sid, result)

        runner = web.AppRunner(self.app)
        await runner.setup()
        site = web.TCPSite(runner, 'localhost', 3000)
        await site.start()
        print("Websocket server started at http://localhost:3000")
        while True:
            await asyncio.sleep(3600)  # Keeps the server running
        server_url = "Websocket server started at http://localhost:3000"
        print(server_url)
        return server_url



# from .hugging_face_model import HuggingFaceModel
# from .google_vertex_ai_model import GoogleVertexAIModel
from .openai_model import OpenAIModel

# __all__ = ["HuggingFaceModel", "GoogleVertexAIModel", "OpenAIModel"]
__all__ = ["OpenAIModel"]



import os
from pathlib import Path

class ChatAction:
    def __init__(self, agent):
        self.dependencies = []
        self.agent = agent
        self.name = "chat"
        self.description = "Chat with agent on a browser."
        self.parameters = []

    async def run(self, args):
        # Resolve the path to the HTML file
        filename = Path(__file__).parent.parent / 'islands/chat.html'
        # Read the HTML content from the file
        with open(filename, "r", encoding="utf8") as file:
            html_content = file.read()

        # Start the WebSocket server with the HTML content
        response = await self.agent.functions['websocket_server'].run({'htmlContent': html_content})
        return f"Chat Server started: {response}"


from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Union

class PredictionRequest:
    def __init__(self, prompt: Optional[str] = None, max_tokens: Optional[int] = None,
                 temperature: Optional[float] = None, top_p: Optional[float] = None,
                 model: Optional[str] = None, meta: Optional[Dict[str, Any]] = None, **kwargs):
        self.prompt = prompt
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.top_p = top_p
        self.model = model
        self.meta = meta
        self.extra_args = kwargs

class PredictionResponse:
    def __init__(self, text: Union[str, Any], model: str, other_metadata: Optional[Any] = None):
        self.text = text
        self.model = model
        self.other_metadata = other_metadata

class LLM(ABC):
    @abstractmethod
    def interact(self, use_delegate: Optional[bool] = False) -> Union[str, None]:
        pass

    @abstractmethod
    def predict(self, request: PredictionRequest) -> PredictionResponse:
        pass


import os
import subprocess
import asyncio
import tempfile
from abc import ABC, abstractmethod
import queue
import threading
import time
import traceback

class LanguageRunner(ABC):
    @abstractmethod
    async def run_code(self, code: str):
        pass

class GeneralRunner(LanguageRunner):
    async def run_code(self, command: str):
        process = await asyncio.create_subprocess_shell(
            command, 
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = await process.communicate()
        if process.returncode == 0:
            return f"Execution complete. {stdout.decode()}"
        else:
            raise Exception(f"Exit with code: {process.returncode}\nError Output:\n{stderr.decode()}")

class PythonRunner(LanguageRunner):
    async def run_code(self, code: str):
        # Create a temporary file
        with tempfile.NamedTemporaryFile(mode='w+', suffix='.py', delete=False) as temp_file:
            temp_file.write(code)
            temp_file_path = temp_file.name

        try:
            # Execute the temporary file
            process = await asyncio.create_subprocess_exec(
                'python3', temp_file_path,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE
            )
            stdout, stderr = await process.communicate()

            if process.returncode != 0:
                raise Exception(f"Script exited with code {process.returncode}\nError Output:\n{stderr.decode()}")

            return f"Output:\n{stdout.decode()}"
        finally:
            # Delete the temporary file
            os.remove(temp_file_path)


class ShellRunner(LanguageRunner):
    def __init__(self):
        self.output_queue = queue.Queue()
        self.done = threading.Event()

    def start_process(self, cmd):
        my_env = os.environ.copy()
        my_env["PYTHONIOENCODING"] = "utf-8"
        process = subprocess.Popen(
            cmd,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=0,
            universal_newlines=True,
            env=my_env,
        )
        threading.Thread(
            target=self.handle_stream_output,
            args=(process.stdout, False),
            daemon=True,
        ).start()
        threading.Thread(
            target=self.handle_stream_output,
            args=(process.stderr, True),
            daemon=True,
        ).start()
        return process

    async def run_code(self, code):
        output_lines = []
        try:
            process = self.start_process(code)
        except:
            return traceback.format_exc()

        while True:
            try:
                output = self.output_queue.get(timeout=0.3)  # Waits for 0.3 seconds
                output_lines.append(output["output"])
            except queue.Empty:
                if self.done.is_set():
                    break
                time.sleep(0.1)

        process.terminate()
        return ''.join(output_lines)

    def handle_stream_output(self, stream, is_error_stream):
        for line in iter(stream.readline, ""):
            self.output_queue.put({"output": line})
            print(line, end="")
        self.done.set()
        
class AppleScriptRunner(LanguageRunner):
    async def run_code(self, code: str):
        process = await asyncio.create_subprocess_shell(
            f"osascript -e \"{code}\"", 
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = await process.communicate()
        if process.returncode == 0:
            return stdout.decode()
        else:
            raise Exception(f"Exit with code: {process.returncode}\nError Output:\n{stderr.decode()}")


class ExecuteCodeAction:
    def __init__(self, agent):
        self.agent = agent
        self.dependencies = ["asyncio"]
        self.name = "execute_code"
        self.description = "Execute code in a specific language"
        self.parameters = [
            {
                "name": "language",
                "type": "string",
                "required": True,
                "enum": ["python", "shell", "bash", "applescript", "AppleScript"]
            },
            { 
                "name": "code", 
                "type": "string", 
                "required": True 
            }
        ]
        self.runner_mapping = {
            "python": PythonRunner(),
            "shell": ShellRunner(),
            "bash": ShellRunner(),
            "applescript": AppleScriptRunner(),
            "AppleScript": AppleScriptRunner(),
            # Other mappings as required
        }

    async def run(self, args):
        language = args.get("language")
        code = args.get("code")
        if language not in self.parameters[0]["enum"]:
            raise ValueError(f"Unsupported language: {language}")

        runner = self.runner_mapping.get(language, GeneralRunner())

        if not isinstance(runner, LanguageRunner):
            raise ValueError(f"Invalid runner for language: {language}")

        try:
            output = await runner.run_code(code)
            return f"output is: {output}"
        except Exception as e:
            error_info = {"message": str(e)}
            print(f"Error occurred: {error_info}")
            return str(error_info)




# llms/openai_model.py

import json
import os
import sys
from openai import OpenAI
from typing import Any, Dict, Optional, Union
from ..interfaces.llm import LLM, PredictionRequest, PredictionResponse

openai = OpenAI()

class OpenAIPredictionRequest(PredictionRequest):
    def __init__(self, model: str, messages: list, max_tokens: Optional[int] = None,
                 temperature: Optional[float] = None, top_p: Optional[float] = None,
                 tools: Optional[Any] = None, **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.messages = messages
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.top_p = top_p
        self.tools = tools

class OpenAIPredictionResponse(PredictionResponse):
    def __init__(self, text: str, model: str, message: Any, other_metadata: Optional[Any] = None):
        super().__init__(text, model, other_metadata)
        self.message = message

class OpenAIModel(LLM):
    def __init__(self, agent, opts: Dict[str, Optional[str]]):
        self.agent = agent
        self.api_key = opts.get("apiKey", os.environ.get("OPENAI_API_KEY", ""))
        
        self.name = os.environ.get("OPENAI_MODEL", "gpt-4-1106-preview")
        self.messages = [
            {
                "role": "system",
                "content": agent.system_message or "You are a helpful assistant",
            },
        ]

    def predict(self, request):
        try:
            # Remove 'prompt' key from the request if it exists
            filtered_request = {k: v for k, v in request.items() if k != 'prompt'}
            model = request.get("model", "gpt-4-1106-preview")
            
            # Make an asynchronous call to OpenAI API
            response = openai.chat.completions.create(**filtered_request)
            if response and hasattr(response, 'choices') and response.choices:
                choice = response.choices[0].message
                tool_calls = getattr(choice, 'tool_calls', [])
                content = getattr(choice, 'content', '')

                text = tool_calls if tool_calls else content or ""

                return OpenAIPredictionResponse(text=text, model=model, message=choice, other_metadata=None)
            else:
                return OpenAIPredictionResponse(text="", model= model, message=None, other_metadata=None)

        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            print(f"An error occurred: {e}")
            raise

    async def interact(self, use_delegate: bool = False) -> Union[str, None]:
        decision = await self.agent.think()
        if isinstance(decision.text, list):
            tool_calls = decision.text
        else:
            tool_calls = []

        content = decision.text if isinstance(decision.text, str) else None
        self.agent.messages.append(decision.message)

        if content:
            if use_delegate:
                return content
            else:
                if "both" in self.agent.options['speech'] or "output" in self.agent.options['speech']:
                    await self.agent.speak(content)
                self.agent.display_message(content)
        else:
            for tool_call in tool_calls:
                action_name = tool_call.function.name if tool_call.function and tool_call.function.name else ""
                args = tool_call.function.arguments if tool_call.function and tool_call.function.arguments else ""
                result = ""
                if (self.agent.memory.last_action == action_name and 
                    self.agent.memory.last_action_status == "failure"):
                    continue  # Skip the repeated action if it previously failed

                try:
                    args = json.loads(args)
                    if not self.agent.options["allow_code_execution"]:
                        # Prompt logic for execution confirmation
                        # This might be different in Python. You might need to use an alternative to 'prompts'
                        answer = input(f"Do you want to execute the code? (y/n): ").lower() == 'y'
                        if not answer:
                            result = "Code execution cancelled for current action only"
                        else:
                            result = await self.agent.act(action_name, args)
                    else:
                        result = await self.agent.act(action_name, args)
                except Exception as e:
                    print(f"An error occurred: {e}")
                    result = str(e)

                self.agent.messages.append({
                    "tool_call_id": tool_call.id,
                    "role": "tool",
                    "name": action_name,
                    "content": result
                })

            return await self.interact(use_delegate)


import os
import cv2
import base64
import requests
import youtube_dl
import openai
from pathlib import Path
from typing import List

class VisionAction:
    def __init__(self, agent):
        self.agent = agent
        self.name = 'openai_vision'
        self.description = 'Analyze a video or image using OpenAI Vision.'
        self.parameters = [
            {'name': 'source', 'type': 'string', 'required': True,
             'description': 'Local path or URL of the video/image.'},
            {'name': 'request', 'type': 'string', 'required': True,
             'description': 'The user request to be sent to OpenAI Vision.'}
        ]

    async def run(self, args: dict) -> str:
        source = args['source']
        openai_request = args['request']
        is_url = source.startswith('http://') or source.startswith('https://')
        file_path = source if not is_url else self.download_media(source)

        if file_path.lower().endswith(('.png', '.jpg', '.jpeg')):
            base64_frames = [self.encode_image_to_base64(file_path)]
        else:
            frames_path = 'tmp/video/frames'
            os.makedirs(frames_path, exist_ok=True)
            self.extract_frames(file_path, frames_path)
            base64_frames = self.encode_frames_to_base64(frames_path)

        return await self.analyze_media(base64_frames, openai_request)

    def download_media(self, url: str) -> str:
        if 'youtube.com' in url or 'youtu.be' in url:
            os.makedirs('tmp/videos', exist_ok=True)
            ydl_opts = {'outtmpl': 'tmp/videos/%(title)s.%(ext)s'}
            with youtube_dl.YoutubeDL(ydl_opts) as ydl:
                ydl.download([url])
                info = ydl.extract_info(url, download=False)
                return ydl.prepare_filename(info)
        else:
            response = requests.get(url)
            response.raise_for_status()
            os.makedirs('tmp/downloads', exist_ok=True)
            file_path = f'tmp/downloads/{Path(url).name}'
            with open(file_path, 'wb') as file:
                file.write(response.content)
            return file_path

    def extract_frames(self, video_path: str, output_path: str):
        cap = cv2.VideoCapture(video_path)
        frame_count = 0
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            cv2.imwrite(f'{output_path}/frame_{frame_count:03d}.jpg', frame)
            frame_count += 1
        cap.release()

    def encode_image_to_base64(self, image_path: str) -> str:
        with open(image_path, 'rb') as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def encode_frames_to_base64(self, output_path: str) -> List[str]:
        return [self.encode_image_to_base64(f'{output_path}/{f}')
                for f in os.listdir(output_path) if f.endswith('.jpg')]

    async def analyze_media(self, base64_frames: List[str], openai_request: str) -> str:
        

        prompt_messages = [
            {
                "role": "user",
                "content": [
                    openai_request,  # Using the openai_request string as part of the prompt
                    *map(lambda x: {"image": x, "resize": 768}, base64_frames[::50]),
                ],
            },
        ]

        params = {
            "model": "gpt-4-vision-preview",
            "messages": prompt_messages,
            "max_tokens": 200,
        }

        response = openai.chat.completions.create(**params)
        return response.choices[0].message.content



from datetime import datetime
import os
import json
import importlib.util
import pathlib
import platform
import sys
from markdown import markdown
import psutil
from rich.console import Console
from rich.markdown import Markdown
from dotenv import load_dotenv
from ..llms import OpenAIModel
import pygame

class AttrDict(dict):
    def __init__(self, **entries):
        super().__init__(entries)

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"'AttrDict' object has no attribute '{item}'")

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        try:
            del self[item]
        except KeyError:
            raise AttributeError(f"'AttrDict' object has no attribute '{item}'")

        
class Agent:
    def __init__(self, options):
        load_dotenv()
        self.options = {**self.default_options(), **options}
        self.system_message = self.options.get('system_message', 'You are a helpful assistant')
        self.score = 100
        self.messages = []
        self.memory = AttrDict(last_action=None, last_action_status=None)
        self.objectives = []
        self.current_objective = None
        self.current_messages = []
        self.services = {}
        self.functions = {}
        self.init(self.options)
        self.load_all_functions(self.options['actions_path'])
        self.actions = self.get_functions_definitions()

    def default_options(self):
        return {
            'actions_path': "../actions",
            'llm': "OpenAI"
        }

    def init(self, options):
        self.options = options
        llm = self.options.get('llm')
        
        if llm == 'openai':
            self.model = OpenAIModel(self, {
                'apiKey': os.environ.get('OPENAI_API_KEY')
            })
        # elif llm == 'vertexai':
        #     self.model = GoogleVertexAI(self, {
        #         'projectId': os.environ.get('GOOGLE_PROJECT_ID'),
        #         'apiEndpoint': os.environ.get('GOOGLE_API_ENDPOINT'),
        #         'modelId': os.environ.get('GOOGLE_MODEL_ID')
        #     })
        # elif llm == 'ollama':
        #     self.model = Ollama(self, {
        #         'baseURL': os.environ.get('OLLAMA_BASE_URL'),
        #         'model': os.environ.get('OLLAMA_MODEL')
        #     })
        # elif llm == 'huggingface':
        #     self.model = HuggingFace(self, {
        #         'apiKey': os.environ.get('HUGGINGFACE_API_KEY'),
        #         'model': os.environ.get('HUGGINGFACE_MODEL')
        #     })
        # elif llm == 'socket':
        #     self.model = SocketAdapterModel(self, self.options)
        else:
            self.model = OpenAIModel(self, {
                'apiKey': os.environ.get('OPENAI_API_KEY')
            })

    async def listen(self):
        """
        Asynchronously listen to speech and return the converted text.
        """
        if "speech_to_text" in self.functions:
            try:
                return await self.functions["speech_to_text"].run({})
            except Exception as e:
                print(f"Error in speech_to_text function: {e}")
                return ""
        else:
            print("speech_to_text function is not defined")
            return ""

    async def think(self, use_function_calls=True):
        """
        Asynchronously process messages and make a decision using the model.
        """
        try:
            # Prepare the system message
            system_message = {
                "role": "system",
                "content": f"{self.system_message}\n{json.dumps(await self.sense())}"
            }

            # Update the list of messages
            messages = [system_message] + self.messages
            self.current_messages = messages

            # Retrieve the last message from a user
            user_message = next(
                (message for message in reversed(self.current_messages) 
                if isinstance(message, dict) and message.get('role') == 'user'), None
            )
            # user_message = next((message for message in reversed(self.current_messages) if message['role'] == 'user'), None)
            user_message_content = user_message['content'] if user_message else None

            # Limit the number of messages to the last 10, if necessary
            limited_messages = self.current_messages[-10:] if len(self.current_messages) > 10 else self.current_messages

            # Prepare parameters for the model's predict method
            predict_params = {
                "prompt": user_message_content,
                "messages": limited_messages,
                "model": getattr(self.model, "name", None)
            }

            if use_function_calls:
                predict_params.update({"tools": self.actions, "tool_choice": "auto"})

            # Make a decision using the model
            decision = self.model.predict(predict_params)
            return decision

        except Exception as error:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            return str(error)

    async def speak(self, text, use_local=False):
        """
        Asynchronously convert text to speech and play it.
        """
        if not use_local:
            # Requesting AI model to generate speech-friendly text
            response = await self.model.predict({
                'messages': [
                    {'role': 'system', 'content': 'Generate a Siri-friendly speech from the following text, capturing all key points while omitting or rephrasing unsuitable content.'},
                    {'role': 'user', 'content': text}
                ],
                'model': os.environ.get('OPENAI_MODEL', 'gpt-4-1106-preview'),
                'max_tokens': 64,
                'temperature': 0.8
            })
            text = response.get('text', text)

        if platform.system() == 'Darwin':
            # On macOS, use `say` command for speech synthesis
            return await self.functions["execute_code"].run({'code': f'say "{text}"', 'language': 'applescript'})
        else:
            # Use a text-to-speech function for other platforms
            filename = await self.functions["text_to_speech"].run({'text': text})
            # Play the audio file
            # Initialize pygame mixer
            pygame.mixer.init()
            pygame.mixer.music.load(filename)
            pygame.mixer.music.play()
            
            # Wait for the music to finish playing
            while pygame.mixer.music.get_busy():
                pygame.time.Clock().tick(10)

    def display_message(self, message):
        """
        Display a message in the terminal, rendering Markdown content.
        """
        console = Console()
        md = Markdown(message)
        console.print(md)

    def load_all_functions(self, actions_path):
        """
        Load all action modules from the specified directory.
        """
        actions_dir = pathlib.Path(__file__).parent.resolve() / actions_path
        for file_path in actions_dir.glob('*.py'):
            if file_path.name == "__init__.py":
                continue

            module_name = file_path.stem
            spec = importlib.util.spec_from_file_location(module_name, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            action_class_name = ''.join(word.title() for word in module_name.split('_')) + 'Action'
            action_class = getattr(module, action_class_name, None)
            if action_class:
                action_instance = action_class(self)
                self.functions[action_instance.name] = action_instance

    def load_functions(self, actions_path):
        """
        Load specific action modules from the specified directory based on the .saiku configuration.
        """
        actions_dir = pathlib.Path(__file__).parent.resolve() / actions_path

        # Default activated actions
        activated_actions = ['execute_code', 'chat', 'websocket_server']

        # Load additional activated actions from .saiku file if it exists
        saiku_file_path = actions_dir / 'saiku'
        if saiku_file_path.exists():
            with open(saiku_file_path, 'r', encoding='utf-8') as file:
                saiku = json.load(file)
                activated_actions.extend(saiku.get('activatedActions', []))

        # Iterate through each Python file in the directory
        for file_path in actions_dir.glob('*.py'):
            if file_path.name == "__init__.py":
                continue

            # Dynamically import the module
            module_name = file_path.stem  # Extracts module name without '.py'
            spec = importlib.util.spec_from_file_location(module_name, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Instantiate the action class if it's in the list of activated actions
            action_class_name = ''.join(word.title() for word in module_name.split('_')) + 'Action'
            action_class = getattr(module, action_class_name, None)
            if action_class:
                action_instance = action_class(self)
                if action_instance.name in activated_actions:
                    self.functions[action_instance.name] = action_instance

    def get_all_functions(self):
        """
        Load and return all action instances from the specified directory.
        """
        actions_dir = pathlib.Path(__file__).parent.resolve() / self.options['actions_path']
        functions = []

        # Iterate through each Python file in the directory
        for file_path in actions_dir.glob('*.py'):
            if file_path.name == "__init__.py":
                continue

            # Dynamically import the module
            module_name = file_path.stem  # Extracts module name without '.py'
            spec = importlib.util.spec_from_file_location(module_name, file_path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Instantiate the action class and add it to the list
            action_class = getattr(module, 'Action', None)
            if action_class:
                action_instance = action_class(self)
                functions.append(action_instance)

        return functions

    async def sense(self):
        """
        Gather and return various system and environment information.
        """
        # System and environment information
        system_info = {
            "agent": {
                "name": "Saiku"
            },
            "os": platform.system(),
            "arch": platform.machine(),
            "version": platform.version(),
            "memory": {
                "total": psutil.virtual_memory().total,
                "used": psutil.virtual_memory().used
            },
            "cpu": {
                "percent": psutil.cpu_percent(interval=1)
            },
            "uptime": psutil.boot_time(),
            "date": datetime.now().strftime("%Y-%m-%d"),
            "start_time": datetime.now().strftime("%H:%M:%S"),
            "cwd": os.getcwd(),
            "current_user": {
                "name": os.environ.get("ME"),
                "country": os.environ.get("COUNTRY"),
                "city": os.environ.get("CITY"),
                "company": os.environ.get("COMPANY"),
                "phone": os.environ.get("PHONE")
            },
            "api_services": {
                "weather": os.environ.get("WEATHER_API_KEY"),
                "gitlab": {
                    "version": os.environ.get("GITLAB_VERSION"),
                    "username": os.environ.get("GITLAB_USERNAME"),
                    "api_version": os.environ.get("GITLAB_API_VERSION")
                }
            },
            **self.memory
        }

        return system_info

    async def act(self, action_name, args):
        """
        Execute a specified action and handle success or failure.
        """
        try:
            action = self.functions.get(action_name)
            self.display_message(f"_Executing action **{action_name}: {getattr(action, 'description', 'No description')}**_")

            if action:
                try:
                    output = await action.run(args)
                    self.update_memory({
                        "": action_name,
                        "last_action_status": "success",
                    })
                    return output
                except Exception as error:
                    self.update_memory({
                        "last_action": action_name,
                        "last_action_status": "failure",
                    })
                    return json.dumps({"error": str(error)})

            else:
                self.display_message(f"No action found with name: **{action_name}**")
                return "Action not found"

        except Exception as error:
            return json.dumps({"error": str(error)})

    def evaluate_performance(self):
        """
        Evaluate the agent's performance based on its objectives.
        Returns a score. For now, this is a placeholder value.
        """
        return self.score

    def remember(self, key, value):
        """
        Store a value in the agent's memory under a specific key.
        """
        self.memory[key] = value

    def recall(self, key):
        """
        Retrieve a value from the agent's memory by key.
        """
        return self.memory.get(key)

    def forget(self, key):
        """
        Remove a value from the agent's memory by key.
        """
        if key in self.memory:
            del self.memory[key]

    def save_memory(self):
        """
        Save the agent's memory to a file.
        """
        memory_file_path = pathlib.Path(__file__).parent / "../data/memory.json"
        with open(memory_file_path, "w") as file:
            json.dump(self.memory, file)

    def get_memory(self):
        """
        Retrieve the agent's current memory state.
        """
        return self.memory

    def update_memory(self, args):
        """
        Update the agent's memory with new information.
        """
        self.memory.update(args)

    async def interact(self, delegate=False):
        """
        Interact with the agent's model.
        """
        return await self.model.interact(delegate)

    def get_functions_definitions(self):
        """
        Get the definitions of all loaded functions.
        """
        actions_definitions = []

        for action in self.functions.values():
            action_def = {
                "name": getattr(action, "name", "Unnamed"),
                "description": getattr(action, "description", ""),
                "parameters": getattr(action, "parameters", [])
            }

            function_def = {
                "type": "function",
                "function": {
                    "name": action_def["name"],
                    "description": action_def["description"],
                    "parameters": self._format_parameters(action_def["parameters"])
                }
            }

            actions_definitions.append(function_def)

        return actions_definitions

    def _format_parameters(self, parameters):
        """
        Format the parameters of an action for output.
        """
        param_def = {
            "type": "object",
            "properties": {},
            "required": []
        }

        for param in parameters:
            param_def["properties"][param["name"]] = {
                "type": param.get("type", "unknown"),
                "description": param.get("description", "")
            }
            if param.get("type") == "array" and "items" in param:
                param_def["properties"][param["name"]]["items"] = param["items"]

            if param.get("required", False):
                param_def["required"].append(param["name"])

        return param_def


import os
import queue
import openai
import sounddevice as sd
import wavio


class SpeechToTextAction:
    dependencies = ["openai", "sounddevice", "wavio"]

    def __init__(self, agent):
        self.agent = agent
        self.name = "speech_to_text"
        self.description = "Transcribe audio to text"
        self.parameters = [{"name": "audioFilename", "type": "string", "required": False}]
        self.audioFilename = 'recording.wav'

    async def init(self):
        # Initialize any other setup tasks if necessary
        pass

    async def run(self, args):
        audioFilename = args.get("audioFilename")
        await self.init()
        audioFilename = audioFilename or self.audioFilename
        if not audioFilename:
            await self.record_audio(audioFilename)
        transcription = await self.transcribe_audio(audioFilename)
        print('Transcription:', transcription)
        return transcription

    async def record_audio(self, filename):
        def callback(indata, frames, time, status):
            if status:
                print(status, file=sys.stderr)
            q.put(indata.copy())

        with sd.InputStream(samplerate=16000, channels=1, callback=callback):
            print("Recording... Press ENTER to stop.")
            await input()  # Waits for ENTER key press
            print("Finished recording")

        q = queue.Queue()
        wavio.write(filename, q.get(), 16000, sampwidth=2)

    async def transcribe_audio(self, filename):
        
        with open(filename, 'rb') as audio_file:
            
            transcript = openai.Audio.transcriptions.create(
                model="whisper-1",
                file=audio_file
            )
        return transcript['text']

    async def close(self):
        # Clean up resources if necessary
        pass



import os
import requests
from openai import OpenAI
from pathlib import Path

class TextToImageAction:
    def __init__(self, agent):
        self.agent = agent
        self.name = 'text_to_image'
        self.description = 'Generates an image based on a description using OpenAI\'s DALL-E 3 model, saves it to disk, and returns the filename.'
        self.parameters = [{
            'name': 'description',
            'type': 'string',
            'required': True,
            'description': self.description
        }]
        dependencies = ['openai', 'requests', 'pathlib']

    async def run(self, description):
        openai = OpenAI(api_key=os.getenv('OPENAI_API_KEY'))

        response = await openai.images.generate(
            model="dall-e-3",
            prompt=description,
            n=1,
            size="1024x1024"
        )

        if not response.data:
            return 'Image generation failed'

        image_url = response.data[0].url
        if not image_url:
            return 'Image generation failed'

        # Download and save the image
        image_response = requests.get(image_url)
        if image_response.status_code == 200:
            # Define the path where you want to save the image
            image_path = Path("generated_images") / f"{description.replace(' ', '_')}.png"
            image_path.parent.mkdir(parents=True, exist_ok=True)

            with open(image_path, "wb") as file:
                file.write(image_response.content)
            
            return f'Image saved to {image_path}'
        else:
            return 'Failed to download the image'